#include "SDL2/SDL.h"

extern SDL_Color pink;
extern SDL_Color purple;
extern SDL_Color blue;
extern SDL_Color green;
extern SDL_Color yellow;
extern SDL_Color orange;
extern SDL_Color lightGrey;
extern SDL_Color grey;
extern SDL_Color darkGrey;
extern SDL_Color black;
extern SDL_Color white;
extern SDL_Color red;